/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "MessageNames.h"
#include "NotificationManagerMessageHandlerMessagesReplies.h"
#include <wtf/Forward.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/UUID.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
struct NotificationData;
}

namespace Messages {
namespace NotificationManagerMessageHandler {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::NotificationManagerMessageHandler;
}

class RequestSystemNotificationPermission {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::NotificationManagerMessageHandler_RequestSystemNotificationPermission; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(bool&&)>&&);
    static void cancelReply(CompletionHandler<void(bool&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::NotificationManagerMessageHandler_RequestSystemNotificationPermissionReply; }
    using AsyncReply = RequestSystemNotificationPermissionAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool allowed);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit RequestSystemNotificationPermission(const String& originIdentifier)
        : m_arguments(originIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ShowNotification {
public:
    using Arguments = std::tuple<const WebCore::NotificationData&>;

    static IPC::MessageName name() { return IPC::MessageName::NotificationManagerMessageHandler_ShowNotification; }
    static constexpr bool isSync = false;

    explicit ShowNotification(const WebCore::NotificationData& notificationData)
        : m_arguments(notificationData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CancelNotification {
public:
    using Arguments = std::tuple<const UUID&>;

    static IPC::MessageName name() { return IPC::MessageName::NotificationManagerMessageHandler_CancelNotification; }
    static constexpr bool isSync = false;

    explicit CancelNotification(const UUID& notificationID)
        : m_arguments(notificationID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ClearNotifications {
public:
    using Arguments = std::tuple<const Vector<UUID>&>;

    static IPC::MessageName name() { return IPC::MessageName::NotificationManagerMessageHandler_ClearNotifications; }
    static constexpr bool isSync = false;

    explicit ClearNotifications(const Vector<UUID>& notificationIDs)
        : m_arguments(notificationIDs)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidDestroyNotification {
public:
    using Arguments = std::tuple<const UUID&>;

    static IPC::MessageName name() { return IPC::MessageName::NotificationManagerMessageHandler_DidDestroyNotification; }
    static constexpr bool isSync = false;

    explicit DidDestroyNotification(const UUID& notificationID)
        : m_arguments(notificationID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace NotificationManagerMessageHandler
} // namespace Messages
