/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "MessageNames.h"
#include "WebIDBServerMessagesReplies.h"
#include <WebCore/IndexedDB.h>
#include <wtf/Forward.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class IDBCursorInfo;
class IDBIndexInfo;
class IDBKeyData;
class IDBObjectStoreInfo;
class IDBRequestData;
class IDBResourceIdentifier;
class IDBTransactionInfo;
class IDBValue;
struct ClientOrigin;
struct IDBGetAllRecordsData;
struct IDBGetRecordData;
struct IDBIterateCursorData;
struct IDBKeyRangeData;
}

namespace Messages {
namespace WebIDBServer {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::WebIDBServer;
}

class DeleteDatabase {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_DeleteDatabase; }
    static constexpr bool isSync = false;

    explicit DeleteDatabase(const WebCore::IDBRequestData& requestData)
        : m_arguments(requestData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class OpenDatabase {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_OpenDatabase; }
    static constexpr bool isSync = false;

    explicit OpenDatabase(const WebCore::IDBRequestData& requestData)
        : m_arguments(requestData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AbortTransaction {
public:
    using Arguments = std::tuple<const WebCore::IDBResourceIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_AbortTransaction; }
    static constexpr bool isSync = false;

    explicit AbortTransaction(const WebCore::IDBResourceIdentifier& transactionIdentifier)
        : m_arguments(transactionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CommitTransaction {
public:
    using Arguments = std::tuple<const WebCore::IDBResourceIdentifier&, uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_CommitTransaction; }
    static constexpr bool isSync = false;

    CommitTransaction(const WebCore::IDBResourceIdentifier& transactionIdentifier, uint64_t pendingRequestCount)
        : m_arguments(transactionIdentifier, pendingRequestCount)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidFinishHandlingVersionChangeTransaction {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::IDBResourceIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_DidFinishHandlingVersionChangeTransaction; }
    static constexpr bool isSync = false;

    DidFinishHandlingVersionChangeTransaction(uint64_t databaseConnectionIdentifier, const WebCore::IDBResourceIdentifier& transactionIdentifier)
        : m_arguments(databaseConnectionIdentifier, transactionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CreateObjectStore {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBObjectStoreInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_CreateObjectStore; }
    static constexpr bool isSync = false;

    CreateObjectStore(const WebCore::IDBRequestData& requestData, const WebCore::IDBObjectStoreInfo& info)
        : m_arguments(requestData, info)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteObjectStore {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_DeleteObjectStore; }
    static constexpr bool isSync = false;

    DeleteObjectStore(const WebCore::IDBRequestData& requestData, const String& objectStoreName)
        : m_arguments(requestData, objectStoreName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RenameObjectStore {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_RenameObjectStore; }
    static constexpr bool isSync = false;

    RenameObjectStore(const WebCore::IDBRequestData& requestData, uint64_t objectStoreIdentifier, const String& newName)
        : m_arguments(requestData, objectStoreIdentifier, newName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ClearObjectStore {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_ClearObjectStore; }
    static constexpr bool isSync = false;

    ClearObjectStore(const WebCore::IDBRequestData& requestData, uint64_t objectStoreIdentifier)
        : m_arguments(requestData, objectStoreIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CreateIndex {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBIndexInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_CreateIndex; }
    static constexpr bool isSync = false;

    CreateIndex(const WebCore::IDBRequestData& requestData, const WebCore::IDBIndexInfo& info)
        : m_arguments(requestData, info)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteIndex {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_DeleteIndex; }
    static constexpr bool isSync = false;

    DeleteIndex(const WebCore::IDBRequestData& requestData, uint64_t objectStoreIdentifier, const String& indexName)
        : m_arguments(requestData, objectStoreIdentifier, indexName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class RenameIndex {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, uint64_t, uint64_t, const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_RenameIndex; }
    static constexpr bool isSync = false;

    RenameIndex(const WebCore::IDBRequestData& requestData, uint64_t objectStoreIdentifier, uint64_t indexIdentifier, const String& newName)
        : m_arguments(requestData, objectStoreIdentifier, indexIdentifier, newName)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class PutOrAdd {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBKeyData&, const WebCore::IDBValue&, const WebCore::IndexedDB::ObjectStoreOverwriteMode&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_PutOrAdd; }
    static constexpr bool isSync = false;

    PutOrAdd(const WebCore::IDBRequestData& requestData, const WebCore::IDBKeyData& key, const WebCore::IDBValue& value, const WebCore::IndexedDB::ObjectStoreOverwriteMode& overwriteMode)
        : m_arguments(requestData, key, value, overwriteMode)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetRecord {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBGetRecordData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_GetRecord; }
    static constexpr bool isSync = false;

    GetRecord(const WebCore::IDBRequestData& requestData, const WebCore::IDBGetRecordData& getRecordData)
        : m_arguments(requestData, getRecordData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetAllRecords {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBGetAllRecordsData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_GetAllRecords; }
    static constexpr bool isSync = false;

    GetAllRecords(const WebCore::IDBRequestData& requestData, const WebCore::IDBGetAllRecordsData& getAllRecordsData)
        : m_arguments(requestData, getAllRecordsData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetCount {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBKeyRangeData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_GetCount; }
    static constexpr bool isSync = false;

    GetCount(const WebCore::IDBRequestData& requestData, const WebCore::IDBKeyRangeData& range)
        : m_arguments(requestData, range)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DeleteRecord {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBKeyRangeData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_DeleteRecord; }
    static constexpr bool isSync = false;

    DeleteRecord(const WebCore::IDBRequestData& requestData, const WebCore::IDBKeyRangeData& range)
        : m_arguments(requestData, range)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class OpenCursor {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBCursorInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_OpenCursor; }
    static constexpr bool isSync = false;

    OpenCursor(const WebCore::IDBRequestData& requestData, const WebCore::IDBCursorInfo& info)
        : m_arguments(requestData, info)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class IterateCursor {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&, const WebCore::IDBIterateCursorData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_IterateCursor; }
    static constexpr bool isSync = false;

    IterateCursor(const WebCore::IDBRequestData& requestData, const WebCore::IDBIterateCursorData& data)
        : m_arguments(requestData, data)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class EstablishTransaction {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::IDBTransactionInfo&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_EstablishTransaction; }
    static constexpr bool isSync = false;

    EstablishTransaction(uint64_t databaseConnectionIdentifier, const WebCore::IDBTransactionInfo& info)
        : m_arguments(databaseConnectionIdentifier, info)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DatabaseConnectionPendingClose {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_DatabaseConnectionPendingClose; }
    static constexpr bool isSync = false;

    explicit DatabaseConnectionPendingClose(uint64_t databaseConnectionIdentifier)
        : m_arguments(databaseConnectionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DatabaseConnectionClosed {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_DatabaseConnectionClosed; }
    static constexpr bool isSync = false;

    explicit DatabaseConnectionClosed(uint64_t databaseConnectionIdentifier)
        : m_arguments(databaseConnectionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class AbortOpenAndUpgradeNeeded {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::IDBResourceIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_AbortOpenAndUpgradeNeeded; }
    static constexpr bool isSync = false;

    AbortOpenAndUpgradeNeeded(uint64_t databaseConnectionIdentifier, const WebCore::IDBResourceIdentifier& transactionIdentifier)
        : m_arguments(databaseConnectionIdentifier, transactionIdentifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidFireVersionChangeEvent {
public:
    using Arguments = std::tuple<uint64_t, const WebCore::IDBResourceIdentifier&, const WebCore::IndexedDB::ConnectionClosedOnBehalfOfServer&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_DidFireVersionChangeEvent; }
    static constexpr bool isSync = false;

    DidFireVersionChangeEvent(uint64_t databaseConnectionIdentifier, const WebCore::IDBResourceIdentifier& requestIdentifier, const WebCore::IndexedDB::ConnectionClosedOnBehalfOfServer& connectionClosedOnBehalfOfServer)
        : m_arguments(databaseConnectionIdentifier, requestIdentifier, connectionClosedOnBehalfOfServer)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class OpenDBRequestCancelled {
public:
    using Arguments = std::tuple<const WebCore::IDBRequestData&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_OpenDBRequestCancelled; }
    static constexpr bool isSync = false;

    explicit OpenDBRequestCancelled(const WebCore::IDBRequestData& requestData)
        : m_arguments(requestData)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetAllDatabaseNamesAndVersions {
public:
    using Arguments = std::tuple<const WebCore::IDBResourceIdentifier&, const WebCore::ClientOrigin&>;

    static IPC::MessageName name() { return IPC::MessageName::WebIDBServer_GetAllDatabaseNamesAndVersions; }
    static constexpr bool isSync = false;

    GetAllDatabaseNamesAndVersions(const WebCore::IDBResourceIdentifier& requestIdentifier, const WebCore::ClientOrigin& origin)
        : m_arguments(requestIdentifier, origin)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

} // namespace WebIDBServer
} // namespace Messages
