/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "WebLockRegistryProxy.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "WebCoreArgumentCoders.h"
#include "WebLockRegistryProxyMessages.h"
#include <WebCore/ClientOrigin.h>
#include <WebCore/ProcessQualified.h>
#include <WebCore/ScriptExecutionContextIdentifier.h>
#include <WebCore/WebLockIdentifier.h>
#include <WebCore/WebLockManagerSnapshot.h>
#include <WebCore/WebLockMode.h>
#include <wtf/ObjectIdentifier.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebLockRegistryProxy {

void AbortLockRequest::callReply(IPC::Decoder& decoder, CompletionHandler<void(bool&&)>&& completionHandler)
{
    std::optional<bool> wasAborted;
    decoder >> wasAborted;
    if (!wasAborted) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*wasAborted));
}

void AbortLockRequest::cancelReply(CompletionHandler<void(bool&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<bool>::create());
}

void AbortLockRequest::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, bool wasAborted)
{
    encoder.get() << wasAborted;
    connection.sendSyncReply(WTFMove(encoder));
}

void Snapshot::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebCore::WebLockManagerSnapshot&&)>&& completionHandler)
{
    std::optional<WebCore::WebLockManagerSnapshot> snapshot;
    decoder >> snapshot;
    if (!snapshot) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*snapshot));
}

void Snapshot::cancelReply(CompletionHandler<void(WebCore::WebLockManagerSnapshot&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebCore::WebLockManagerSnapshot>::create());
}

void Snapshot::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::WebLockManagerSnapshot& snapshot)
{
    encoder.get() << snapshot;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebLockRegistryProxy

} // namespace Messages

namespace WebKit {

void WebLockRegistryProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::WebLockRegistryProxy::RequestLock::name())
        return IPC::handleMessage<Messages::WebLockRegistryProxy::RequestLock>(connection, decoder, this, &WebLockRegistryProxy::requestLock);
    if (decoder.messageName() == Messages::WebLockRegistryProxy::ReleaseLock::name())
        return IPC::handleMessage<Messages::WebLockRegistryProxy::ReleaseLock>(connection, decoder, this, &WebLockRegistryProxy::releaseLock);
    if (decoder.messageName() == Messages::WebLockRegistryProxy::AbortLockRequest::name())
        return IPC::handleMessageAsync<Messages::WebLockRegistryProxy::AbortLockRequest>(connection, decoder, this, &WebLockRegistryProxy::abortLockRequest);
    if (decoder.messageName() == Messages::WebLockRegistryProxy::Snapshot::name())
        return IPC::handleMessageAsync<Messages::WebLockRegistryProxy::Snapshot>(connection, decoder, this, &WebLockRegistryProxy::snapshot);
    if (decoder.messageName() == Messages::WebLockRegistryProxy::ClientIsGoingAway::name())
        return IPC::handleMessage<Messages::WebLockRegistryProxy::ClientIsGoingAway>(connection, decoder, this, &WebLockRegistryProxy::clientIsGoingAway);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled message %s to %" PRIu64, IPC::description(decoder.messageName()), decoder.destinationID());
}

} // namespace WebKit
