/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "WebProcessPool.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "UserData.h"
#include "WebProcessPoolMessages.h"
#include <wtf/Seconds.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebProcessPool {

void HandleSynchronousMessage::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::UserData& returnData)
{
    encoder.get() << returnData;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebProcessPool

} // namespace Messages

namespace WebKit {

void WebProcessPool::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    Ref protectedThis { *this };
    if (decoder.messageName() == Messages::WebProcessPool::HandleMessage::name())
        return IPC::handleMessageWantsConnection<Messages::WebProcessPool::HandleMessage>(connection, decoder, this, &WebProcessPool::handleMessage);
#if ENABLE(GAMEPAD)
    if (decoder.messageName() == Messages::WebProcessPool::StartedUsingGamepads::name())
        return IPC::handleMessageWantsConnection<Messages::WebProcessPool::StartedUsingGamepads>(connection, decoder, this, &WebProcessPool::startedUsingGamepads);
#endif
#if ENABLE(GAMEPAD)
    if (decoder.messageName() == Messages::WebProcessPool::StoppedUsingGamepads::name())
        return IPC::handleMessageWantsConnection<Messages::WebProcessPool::StoppedUsingGamepads>(connection, decoder, this, &WebProcessPool::stoppedUsingGamepads);
#endif
    if (decoder.messageName() == Messages::WebProcessPool::ReportWebContentCPUTime::name())
        return IPC::handleMessage<Messages::WebProcessPool::ReportWebContentCPUTime>(connection, decoder, this, &WebProcessPool::reportWebContentCPUTime);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled message %s to %" PRIu64, IPC::description(decoder.messageName()), decoder.destinationID());
}

bool WebProcessPool::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, UniqueRef<IPC::Encoder>& replyEncoder)
{
    Ref protectedThis { *this };
    if (decoder.messageName() == Messages::WebProcessPool::HandleSynchronousMessage::name())
        return IPC::handleMessageSynchronousWantsConnection<Messages::WebProcessPool::HandleSynchronousMessage>(connection, decoder, replyEncoder, this, &WebProcessPool::handleSynchronousMessage);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return false;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled synchronous message %s to %" PRIu64, description(decoder.messageName()), decoder.destinationID());
    return false;
}

} // namespace WebKit
