/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgumentCoders.h"
#include "Connection.h"
#include "DisplayLinkObserverID.h"
#include "MessageNames.h"
#include "WebProcessProxyMessagesReplies.h"
#include <WebCore/CaptionUserPreferences.h>
#include <WebCore/FrameIdentifier.h>
#include <WebCore/PageIdentifier.h>
#include <WebCore/SleepDisablerIdentifier.h>
#include <wtf/Forward.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {
class RegistrableDomain;
struct PrewarmInformation;
}

namespace WebKit {
struct BackForwardListItemState;
struct GPUProcessConnectionInfo;
struct GPUProcessConnectionParameters;
struct NetworkProcessConnectionInfo;
struct UserMessage;
struct WebAuthnProcessConnectionInfo;
}

namespace Messages {
namespace WebProcessProxy {

static inline IPC::ReceiverName messageReceiverName()
{
    return IPC::ReceiverName::WebProcessProxy;
}

class UpdateBackForwardItem {
public:
    using Arguments = std::tuple<const WebKit::BackForwardListItemState&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_UpdateBackForwardItem; }
    static constexpr bool isSync = false;

    explicit UpdateBackForwardItem(const WebKit::BackForwardListItemState& backForwardListItemState)
        : m_arguments(backForwardListItemState)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidDestroyFrame {
public:
    using Arguments = std::tuple<const WebCore::FrameIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidDestroyFrame; }
    static constexpr bool isSync = false;

    explicit DidDestroyFrame(const WebCore::FrameIdentifier& frameID)
        : m_arguments(frameID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidDestroyUserGestureToken {
public:
    using Arguments = std::tuple<uint64_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidDestroyUserGestureToken; }
    static constexpr bool isSync = false;

    explicit DidDestroyUserGestureToken(uint64_t userGestureTokenID)
        : m_arguments(userGestureTokenID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class ShouldTerminate {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_ShouldTerminate; }
    static constexpr bool isSync = true;

    using DelayedReply = ShouldTerminateDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool shouldTerminate);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class EnableSuddenTermination {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_EnableSuddenTermination; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DisableSuddenTermination {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DisableSuddenTermination; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class GetNetworkProcessConnection {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_GetNetworkProcessConnection; }
    static constexpr bool isSync = true;

    using DelayedReply = GetNetworkProcessConnectionDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const WebKit::NetworkProcessConnectionInfo& connectionInfo);
    using Reply = std::tuple<WebKit::NetworkProcessConnectionInfo&>;
    using ReplyArguments = std::tuple<WebKit::NetworkProcessConnectionInfo>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if ENABLE(GPU_PROCESS)
class GetGPUProcessConnection {
public:
    using Arguments = std::tuple<const WebKit::GPUProcessConnectionParameters&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_GetGPUProcessConnection; }
    static constexpr bool isSync = true;

    using DelayedReply = GetGPUProcessConnectionDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const WebKit::GPUProcessConnectionInfo& connectionInfo);
    using Reply = std::tuple<WebKit::GPUProcessConnectionInfo&>;
    using ReplyArguments = std::tuple<WebKit::GPUProcessConnectionInfo>;
    explicit GetGPUProcessConnection(const WebKit::GPUProcessConnectionParameters& parameters)
        : m_arguments(parameters)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if ENABLE(WEB_AUTHN)
class GetWebAuthnProcessConnection {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_GetWebAuthnProcessConnection; }
    static constexpr bool isSync = true;

    using DelayedReply = GetWebAuthnProcessConnectionDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const WebKit::WebAuthnProcessConnectionInfo& connectionInfo);
    using Reply = std::tuple<WebKit::WebAuthnProcessConnectionInfo&>;
    using ReplyArguments = std::tuple<WebKit::WebAuthnProcessConnectionInfo>;
    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class SetIsHoldingLockedFiles {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_SetIsHoldingLockedFiles; }
    static constexpr bool isSync = false;

    explicit SetIsHoldingLockedFiles(bool isHoldingLockedFiles)
        : m_arguments(isHoldingLockedFiles)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidExceedActiveMemoryLimit {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidExceedActiveMemoryLimit; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidExceedInactiveMemoryLimit {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidExceedInactiveMemoryLimit; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidExceedCPULimit {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidExceedCPULimit; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class StopResponsivenessTimer {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_StopResponsivenessTimer; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidReceiveBackgroundResponsivenessPing {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidReceiveBackgroundResponsivenessPing; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class MemoryPressureStatusChanged {
public:
    using Arguments = std::tuple<bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_MemoryPressureStatusChanged; }
    static constexpr bool isSync = false;

    explicit MemoryPressureStatusChanged(bool isUnderMemoryPressure)
        : m_arguments(isUnderMemoryPressure)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidExceedInactiveMemoryLimitWhileActive {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidExceedInactiveMemoryLimitWhileActive; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidCollectPrewarmInformation {
public:
    using Arguments = std::tuple<const WebCore::RegistrableDomain&, const WebCore::PrewarmInformation&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidCollectPrewarmInformation; }
    static constexpr bool isSync = false;

    DidCollectPrewarmInformation(const WebCore::RegistrableDomain& domain, const WebCore::PrewarmInformation& prewarmInformation)
        : m_arguments(domain, prewarmInformation)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(COCOA)
class CacheMediaMIMETypes {
public:
    using Arguments = std::tuple<const Vector<String>&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_CacheMediaMIMETypes; }
    static constexpr bool isSync = false;

    explicit CacheMediaMIMETypes(const Vector<String>& types)
        : m_arguments(types)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class RequestHighPerformanceGPU {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_RequestHighPerformanceGPU; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(MAC)
class ReleaseHighPerformanceGPU {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_ReleaseHighPerformanceGPU; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if HAVE(CVDISPLAYLINK)
class StartDisplayLink {
public:
    using Arguments = std::tuple<const WebKit::DisplayLinkObserverID&, uint32_t, const unsigned&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_StartDisplayLink; }
    static constexpr bool isSync = false;

    StartDisplayLink(const WebKit::DisplayLinkObserverID& observerID, uint32_t displayID, const unsigned& preferredFramesPerSecond)
        : m_arguments(observerID, displayID, preferredFramesPerSecond)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if HAVE(CVDISPLAYLINK)
class StopDisplayLink {
public:
    using Arguments = std::tuple<const WebKit::DisplayLinkObserverID&, uint32_t>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_StopDisplayLink; }
    static constexpr bool isSync = false;

    StopDisplayLink(const WebKit::DisplayLinkObserverID& observerID, uint32_t displayID)
        : m_arguments(observerID, displayID)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if HAVE(CVDISPLAYLINK)
class SetDisplayLinkPreferredFramesPerSecond {
public:
    using Arguments = std::tuple<const WebKit::DisplayLinkObserverID&, uint32_t, const unsigned&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_SetDisplayLinkPreferredFramesPerSecond; }
    static constexpr bool isSync = false;

    SetDisplayLinkPreferredFramesPerSecond(const WebKit::DisplayLinkObserverID& observerID, uint32_t displayID, const unsigned& preferredFramesPerSecond)
        : m_arguments(observerID, displayID, preferredFramesPerSecond)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK) || PLATFORM(WPE)
class SendMessageToWebContext {
public:
    using Arguments = std::tuple<const WebKit::UserMessage&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_SendMessageToWebContext; }
    static constexpr bool isSync = false;

    explicit SendMessageToWebContext(const WebKit::UserMessage& userMessage)
        : m_arguments(userMessage)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if PLATFORM(GTK) || PLATFORM(WPE)
class SendMessageToWebContextWithReply {
public:
    using Arguments = std::tuple<const WebKit::UserMessage&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_SendMessageToWebContextWithReply; }
    static constexpr bool isSync = false;

    static void callReply(IPC::Decoder&, CompletionHandler<void(WebKit::UserMessage&&)>&&);
    static void cancelReply(CompletionHandler<void(WebKit::UserMessage&&)>&&);
    static IPC::MessageName asyncMessageReplyName() { return IPC::MessageName::WebProcessProxy_SendMessageToWebContextWithReplyReply; }
    using AsyncReply = SendMessageToWebContextWithReplyAsyncReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, const WebKit::UserMessage& replyMessage);
    using Reply = std::tuple<WebKit::UserMessage&>;
    using ReplyArguments = std::tuple<WebKit::UserMessage>;
    explicit SendMessageToWebContextWithReply(const WebKit::UserMessage& userMessage)
        : m_arguments(userMessage)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

class DidCreateSleepDisabler {
public:
    using Arguments = std::tuple<const WebCore::SleepDisablerIdentifier&, const String&, bool>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidCreateSleepDisabler; }
    static constexpr bool isSync = false;

    DidCreateSleepDisabler(const WebCore::SleepDisablerIdentifier& identifier, const String& reason, bool display)
        : m_arguments(identifier, reason, display)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DidDestroySleepDisabler {
public:
    using Arguments = std::tuple<const WebCore::SleepDisablerIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DidDestroySleepDisabler; }
    static constexpr bool isSync = false;

    explicit DidDestroySleepDisabler(const WebCore::SleepDisablerIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class CreateSpeechRecognitionServer {
public:
    using Arguments = std::tuple<const WebCore::PageIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_CreateSpeechRecognitionServer; }
    static constexpr bool isSync = false;

    explicit CreateSpeechRecognitionServer(const WebCore::PageIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class DestroySpeechRecognitionServer {
public:
    using Arguments = std::tuple<const WebCore::PageIdentifier&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_DestroySpeechRecognitionServer; }
    static constexpr bool isSync = false;

    explicit DestroySpeechRecognitionServer(const WebCore::PageIdentifier& identifier)
        : m_arguments(identifier)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

class SystemBeep {
public:
    using Arguments = std::tuple<>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_SystemBeep; }
    static constexpr bool isSync = false;

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};

#if PLATFORM(MAC)
class IsAXAuthenticated {
public:
    using Arguments = std::tuple<const audit_token_t&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_IsAXAuthenticated; }
    static constexpr bool isSync = true;

    using DelayedReply = IsAXAuthenticatedDelayedReply;
    static constexpr auto callbackThread = WTF::CompletionHandlerCallThread::ConstructionThread;
    static void send(UniqueRef<IPC::Encoder>&&, IPC::Connection&, bool authenticated);
    using Reply = std::tuple<bool&>;
    using ReplyArguments = std::tuple<bool>;
    explicit IsAXAuthenticated(const audit_token_t& auditToken)
        : m_arguments(auditToken)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if HAVE(MEDIA_ACCESSIBILITY_FRAMEWORK)
class SetCaptionDisplayMode {
public:
    using Arguments = std::tuple<const WebCore::CaptionUserPreferences::CaptionDisplayMode&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_SetCaptionDisplayMode; }
    static constexpr bool isSync = false;

    explicit SetCaptionDisplayMode(const WebCore::CaptionUserPreferences::CaptionDisplayMode& mode)
        : m_arguments(mode)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

#if HAVE(MEDIA_ACCESSIBILITY_FRAMEWORK)
class SetCaptionLanguage {
public:
    using Arguments = std::tuple<const String&>;

    static IPC::MessageName name() { return IPC::MessageName::WebProcessProxy_SetCaptionLanguage; }
    static constexpr bool isSync = false;

    explicit SetCaptionLanguage(const String& language)
        : m_arguments(language)
    {
    }

    const Arguments& arguments() const
    {
        return m_arguments;
    }

private:
    Arguments m_arguments;
};
#endif

} // namespace WebProcessProxy
} // namespace Messages
