/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "WebSharedWorkerObjectConnection.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "WebCoreArgumentCoders.h"
#include "WebSharedWorkerObjectConnectionMessages.h"
#include <WebCore/ResourceError.h>
#include <WebCore/SharedWorkerObjectIdentifier.h>
#include <WebCore/WorkerFetchResult.h>
#include <WebCore/WorkerOptions.h>
#include <wtf/URLHash.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace WebSharedWorkerObjectConnection {

void FetchScriptInClient::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebCore::WorkerFetchResult&&)>&& completionHandler)
{
    std::optional<WebCore::WorkerFetchResult> fetchResult;
    decoder >> fetchResult;
    if (!fetchResult) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*fetchResult));
}

void FetchScriptInClient::cancelReply(CompletionHandler<void(WebCore::WorkerFetchResult&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebCore::WorkerFetchResult>::create());
}

void FetchScriptInClient::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::WorkerFetchResult& fetchResult)
{
    encoder.get() << fetchResult;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace WebSharedWorkerObjectConnection

} // namespace Messages

namespace WebKit {

void WebSharedWorkerObjectConnection::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    Ref protectedThis { *this };
    if (decoder.messageName() == Messages::WebSharedWorkerObjectConnection::FetchScriptInClient::name())
        return IPC::handleMessageAsync<Messages::WebSharedWorkerObjectConnection::FetchScriptInClient>(connection, decoder, this, &WebSharedWorkerObjectConnection::fetchScriptInClient);
    if (decoder.messageName() == Messages::WebSharedWorkerObjectConnection::NotifyWorkerObjectOfLoadCompletion::name())
        return IPC::handleMessage<Messages::WebSharedWorkerObjectConnection::NotifyWorkerObjectOfLoadCompletion>(connection, decoder, this, &WebSharedWorkerObjectConnection::notifyWorkerObjectOfLoadCompletion);
    if (decoder.messageName() == Messages::WebSharedWorkerObjectConnection::PostExceptionToWorkerObject::name())
        return IPC::handleMessage<Messages::WebSharedWorkerObjectConnection::PostExceptionToWorkerObject>(connection, decoder, this, &WebSharedWorkerObjectConnection::postExceptionToWorkerObject);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled message %s to %" PRIu64, IPC::description(decoder.messageName()), decoder.destinationID());
}

} // namespace WebKit
